import logging
import os

logger = logging.getLogger("Anzar")


if os.getenv("ENV") == "dev":
    logging.basicConfig(
        level=logging.INFO,
        format="[%(levelname)s] %(asctime)s: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )
else:
    logger.addHandler(logging.NullHandler())



from pydantic import BaseModel, EmailStr

from .user import User


class LoginRequest(BaseModel):
    email: EmailStr
    password: str


class RegisterRequest(BaseModel):
    username: str
    email: EmailStr
    password: str


class AuthResponse(BaseModel):
    accessToken: str
    refreshToken: str
    user: User


class JWTTokens(BaseModel):
    accessToken: str
    refreshToken: str
    refreshTokenJti: str | None = None



from collections.abc import Mapping
from dataclasses import dataclass
from enum import Enum, auto

from pydantic import BaseModel

type Header = Mapping[str, str | bytes | None]


class NoType(BaseModel):
    status: str | None


class TokenType(Enum):
    AccessToken = auto()
    RefreshToken = auto()


@dataclass()
class Token:
    value: str
    tokenType: TokenType

    @classmethod
    def new(cls, value: str, tokenType: TokenType):
        return cls(value, tokenType)



from pydantic import BaseModel, ValidationError
from requests import Response

from anzar._utils.errors import Error

from anzar._models.auth import LoginRequest, RegisterRequest


class Validator:
    def construct_register(
        self, username: str, email: str, password: str
    ) -> RegisterRequest | Error:
        try:
            return RegisterRequest(username=username, email=email, password=password)

        except ValidationError as e:
            ctx = e.errors()[0].get("ctx")

            reason: str | None = ctx.get("reason") if ctx else None
            return Error(error=reason or "Data is not validated")

    def construct_login(self, email: str, password: str) -> LoginRequest | Error:
        try:
            return LoginRequest(email=email, password=password)
        except ValidationError as e:
            ctx = e.errors()[0].get("ctx")

            reason: str | None = ctx.get("reason") if ctx else None
            return Error(error=reason or "Data is not validated")

    def validate[T: BaseModel](self, model_type: type[T], res: Response) -> T | Error:
        try:
            return model_type.model_validate(res.json())
        except ValidationError as _:
            return Error(error="")



import os
from typing import Any, override

import requests
# from requests.adapters import HTTPAdapter
# from urllib3.util import Retry

from anzar._models.auth import AuthResponse, JWTTokens
from anzar._utils.config import Config
from anzar._utils.context import Context
from anzar._utils.logger import logger
from anzar._utils.storage import TokenStorage
from anzar._utils.types import Token, TokenType


class HttpInterceptor(requests.Session):
    def __init__(self) -> None:
        super().__init__()
        self.API_URL: str = os.getenv("API_URL", "http://localhost:3000")
        self.ctx: Context = Context()
        self.storage: TokenStorage = TokenStorage()
        self.banned_endpoints: list[str] = ["auth/login", "auth/register"]
        self.refresh_endpoints: list[str] = ["auth/logout", "auth/refreshToken"]

        # Retry logic (optional)
        # retries = Retry(
        #     total=3,
        #     backoff_factor=1,
        #     status_forcelist=[500, 502, 503, 504],
        #     allowed_methods=["GET", "POST", "PUT", "DELETE", "PATCH"],
        # )
        # adapter = HTTPAdapter(max_retries=retries)
        # self.mount("http://", adapter)
        # self.mount("https://", adapter)

    def __extractTokenFromCache(self, tokenType: TokenType) -> Token | None:
        token = self.storage.load(tokenType.name)
        return Token.new(token, tokenType) if token else None

    def __isUrlPartOf(self, url: str | bytes, endpoints: list[str]) -> bool:
        return any(url == f"{self.API_URL}/{endpoint}" for endpoint in endpoints)

    def __get_appropriate_token(self, url: str | bytes) -> Token | None:
        if self.__isUrlPartOf(url, self.banned_endpoints):
            return None

        if self.__isUrlPartOf(url, self.refresh_endpoints):
            return self.__extractTokenFromCache(TokenType.RefreshToken)

        return self.__extractTokenFromCache(TokenType.AccessToken)

    def __handle_auth_response(
        self, url: str | bytes, response: requests.Response
    ) -> None:
        if not response.ok:
            return

        if url == f"{self.API_URL}/auth/refreshToken":
            jwtTokens = JWTTokens.model_validate(response.json())
        elif self.__isUrlPartOf(url, self.banned_endpoints):
            auth_response = AuthResponse.model_validate(response.json())
            jwtTokens = JWTTokens(
                accessToken=auth_response.accessToken,
                refreshToken=auth_response.refreshToken,
            )
        else:
            return

        self.storage.save(jwtTokens)

    def __send_request(
        self, method: str | bytes, url: str | bytes, *args, **kwargs
    ) -> requests.Response:
        token = self.__get_appropriate_token(url)
        context_id = self.ctx.load()
        kwargs["headers"] = Config().headers(
            token=token, context_id=context_id if context_id else ""
        )
        return super().request(method, url, timeout=30, *args, **kwargs)

    @override
    def request(
        self,
        method: str | bytes,
        url: str | bytes,
        _refresh_attempted: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> requests.Response:
        # PRE-REQUEST INTERCEPTION
        logger.info(f"{method} {url}")
        response = self.__send_request(method, url, *args, **kwargs)

        # POST-RESPONSE INTERCEPTION
        logger.info(f"Response: {response.status_code}")
        if (
            response.status_code == 401
            and not self.__isUrlPartOf(url, self.banned_endpoints)
            and not self.__isUrlPartOf(url, self.refresh_endpoints)
            and not _refresh_attempted
        ):
            # header: Content-Type: application/x-www-form-urlencoded
            response = self.request(
                method="POST",
                url=f"{self.API_URL}/auth/refreshToken",
                _refresh_attempted=True,
                *args,
                **kwargs,
            )

            response = self.__send_request(method, url, *args, **kwargs)

        self.__handle_auth_response(url, response)

        return response



from anzar._utils.types import Header, Token, TokenType


class Config:
    def headers(self, token: Token | None, context_id: str) -> Header:
        default_headers = {
            "X-Context-ID": context_id,
            "Content-Type": "application/json",
            "User-Agent": "Anzar-SDK/1.0",
        }
        if token is not None:
            if token.tokenType == TokenType.AccessToken:
                default_headers["authorization"] = f"Bearer {token.value}"
            if token.tokenType == TokenType.RefreshToken:
                default_headers["x-refresh-token"] = f"Bearer {token.value}"

        return default_headers



import os
import sys

from anzar._api.client import HttpClient
from anzar._models.anzar_config import AnzarConfig
from anzar._models.auth import AuthResponse
from anzar._models.user import User
from anzar._utils.context import ContextModel
from anzar._utils.errors import Error
from anzar._utils.storage import TokenStorage
from anzar._utils.types import NoType, TokenType
from anzar._utils.validator import Validator


class AuthManager:
    def __init__(self, httpClient: HttpClient, config: AnzarConfig) -> None:
        self._http_client: HttpClient = httpClient

        self._API_URL: str = os.getenv("API_URL", "http://localhost:3000")
        assert self._API_URL is not None, "Env was unable to load"

        self.__endpoints: dict[str, str] = {
            "context": f"{self._API_URL}/configuration/register_context",
            "health_check": f"{self._API_URL}/health_check",
            "login": f"{self._API_URL}/auth/login",
            "register": f"{self._API_URL}/auth/register",
            "user": f"{self._API_URL}/user",
            "logout": f"{self._API_URL}/auth/logout",
        }

        response = self._register_context(config)
        if isinstance(response, Error):
            sys.exit(response.error)
        self._health_check()

    def _register_context(self, config: AnzarConfig) -> Error | ContextModel:
        url = self.__endpoints["context"]
        return self._http_client.post(url, config, ContextModel)

    def _health_check(self):
        try:
            url = self.__endpoints["health_check"]
            _ = self._http_client.health_check(url)
        except ConnectionError as _:
            sys.exit(1)

    def register(self, username: str, email: str, password: str) -> Error | User:
        req = Validator().construct_register(username, email, password)
        if isinstance(req, Error):
            return req

        url = self.__endpoints["register"]
        response = self._http_client.post(url, req, AuthResponse)

        return response.user if isinstance(response, AuthResponse) else response

    def login(self, email: str, password: str):
        req = Validator().construct_login(email, password)
        if isinstance(req, Error):
            return req

        url = self.__endpoints["login"]
        response = self._http_client.post(url, req, AuthResponse)

        return response.user if isinstance(response, AuthResponse) else response

    def getUser(self):
        url = self.__endpoints["user"]
        response = self._http_client.get(url, User)
        return response

    def logout(self):
        url = self.__endpoints["logout"]
        response = self._http_client.post(url, None, NoType)
        TokenStorage().clear()

        return response

    def isLoggedIn(self):
        return TokenStorage().load(TokenType.AccessToken.name) is not None



from pydantic import BaseModel
import requests
from requests.models import Response

from anzar._api.http_interceptor import HttpInterceptor
from anzar._utils.errors import Error
from anzar._utils.logger import logger
from anzar._utils.validator import Validator


class HttpClient:
    def __init__(self, http_interceptor: HttpInterceptor):
        self.http_interceptor: HttpInterceptor = http_interceptor
        self.accessToken: str | None = None

    def health_check(self, url: str) -> bool:
        try:
            response = requests.get(url)
            return response.status_code in (200, 201)
        except Exception as _:
            logger.error(
                "The server is currently unavailable. Please check the server status or try again later."
            )
            raise ConnectionError("Server is down: Unable to establish connection")

    def get[T: BaseModel](self, url: str, model_type: type[T]) -> T | Error:
        response: Response = self.http_interceptor.get(url)

        if response.status_code in (200, 201):
            return model_type.model_validate(response.json())
        else:
            return Error.model_validate(response.json())

    def post[T: BaseModel](
        self, url: str, data: BaseModel | None, model_type: type[T]
    ) -> T | Error:
        payload = data.model_dump() if data else None
        response: Response = self.http_interceptor.post(url, json=payload)

        if response.status_code in (200, 201):
            return Validator().validate(model_type, response)
        else:
            return Error.model_validate(response.json())



import dotenv

from anzar._auth.authenticator import AuthManager
from anzar._models.anzar_config import (
    AnzarConfig,
    Database,
    DatabaseType,
    EmailAndPassword,
)

from ._api.client import HttpClient
from ._api.http_interceptor import HttpInterceptor

_ = dotenv.load_dotenv()


default_config = AnzarConfig(
    api_url="http://localhost:3000",
    database=Database(
        connection_string="sqlite:memory:",
        db_type=DatabaseType.SQLite,
    ),
    emailAndPassword=EmailAndPassword(
        enable=True,
    ),
)


def AnzarAuth(config: AnzarConfig = default_config) -> AuthManager:
    return AuthManager(
        HttpClient(
            HttpInterceptor(),
        ),
        config,
    )


__all__ = ["AnzarAuth"]



import keyring

from anzar._models.auth import JWTTokens
from anzar._utils.types import TokenType


SERVICE_NAME = "AnzarSDK"

ACCESS_TOKEN = TokenType.AccessToken.name
REFRESH_TOKEN = TokenType.RefreshToken.name


class TokenStorage:
    def save(self, tokens: JWTTokens):
        try:
            keyring.set_password(SERVICE_NAME, ACCESS_TOKEN, tokens.accessToken)
            keyring.set_password(SERVICE_NAME, REFRESH_TOKEN, tokens.refreshToken)
        except Exception as e:
            print(e)

    def load(self, username: str):
        try:
            return keyring.get_password(SERVICE_NAME, username)
        except Exception as e:
            print(e)

    def clear(self):
        keyring.delete_password(SERVICE_NAME, ACCESS_TOKEN)
        keyring.delete_password(SERVICE_NAME, REFRESH_TOKEN)



import keyring
from pydantic import BaseModel


class ContextModel(BaseModel):
    context_id: str


SERVICE_NAME = "AnzarSDK"
USERNAME = "Context"


class Context:
    def save(self, context_id: str):
        try:
            keyring.set_password(SERVICE_NAME, USERNAME, context_id)
        except Exception as e:
            print(e)

    def load(self):
        try:
            return keyring.get_password(SERVICE_NAME, USERNAME)
        except Exception as e:
            print(e)

    def clear(self):
        keyring.delete_password(SERVICE_NAME, USERNAME)



from enum import Enum
from pydantic import BaseModel


class DatabaseType(str, Enum):
    SQLite = "SQLite"
    PostgreSQL = "PostgreSQL"
    MongoDB = "MongoDB"


class Database(BaseModel):
    connection_string: str
    db_type: DatabaseType


class EmailAndPassword(BaseModel):
    enable: bool


class AnzarConfig(BaseModel):
    api_url: str
    database: Database
    emailAndPassword: EmailAndPassword



from ._utils.errors import Error
from ._utils.types import NoType

from anzar._models.user import User
from anzar._models.auth import AuthResponse


__all__ = ["User", "AuthResponse", "NoType", "Error"]



from pydantic import BaseModel, EmailStr

from enum import Enum


class Role(str, Enum):
    Admin = "Admin"
    User = "User"


class User(BaseModel):
    # id: str | None = Field(None, alias="_id")
    _id: str | None
    username: str
    email: EmailStr
    role: Role
    isPremium: bool
    accountLocked: bool
